import platform
from typing import Optional

import gzip
import time
import subprocess
from pathlib import Path

import requests
from rich.progress import Progress

from .utils import get_cache_path

VERSION = "0.1.0"


def get_chisel_download_url() -> str:
    system = platform.system().lower()
    arch = platform.machine().lower()

    return f"https://gitee.com/vivym/chisel-buaa/releases/download/v0.1.0/chisel-buaa_{VERSION}_{system}_{arch}.gz"


class Chisel:
    def __init__(
        self,
        server: str,
        cookie: str,
        local_host: str,
        local_port: int,
        remote_host: str,
        remote_port: int,
        auth_key: Optional[str] = None,
        delay: int = 100,
        quiet: bool = False,
    ) -> None:
        self.server = server
        self.cookie = cookie
        self.local_host = local_host
        self.local_port = local_port
        self.remote_host = remote_host
        self.remote_port = remote_port
        self.auth_key = auth_key
        self.delay = delay
        self.quiet = quiet

    def get_executable(self) -> Path:
        chisel_path = get_cache_path() / f"chisel-buaa-{VERSION}"
        chisel_gz_path = chisel_path.with_suffix(".gz")

        if not chisel_path.exists():
            download_url = get_chisel_download_url()
            print(download_url)

            with requests.get(download_url, stream=True) as r:
                with Progress() as progress:
                    task1 = progress.add_task(
                        "Downloading chisel-buaa",
                        total=int(r.headers["Content-Length"]),
                    )
                    task2 = progress.add_task("Uncompressing chisel-buaa")

                    r.raise_for_status()
                    with open(chisel_gz_path, "wb") as f:
                        for chunk in r.iter_content(chunk_size=8192):
                            if chunk:
                                f.write(chunk)
                                progress.update(task1, advance=len(chunk))

                    with gzip.open(chisel_gz_path, "rb") as f_in:
                        with open(chisel_path, "wb") as f_out:
                            for chunk in f_in:
                                if chunk:
                                    f_out.write(chunk)
                                    progress.update(task2, advance=len(chunk))

                    chisel_gz_path.unlink()
                    chisel_path.chmod(0o755)

        return chisel_path.resolve()

    def run(self):
        chisel_path = self.get_executable()

        cmd = [chisel_path, "client"]
        if not self.quiet:
            cmd.append("-v")

        if self.auth_key:
            cmd.append("--auth")
            cmd.append(self.auth_key)

        cmd.append("--header")
        cmd.append(f"Cookie: {self.cookie}")

        cmd.append(f"{self.server}/?delay={self.delay}&t={time.time()}")

        cmd.append(f"{self.local_host}:{self.local_port}:{self.remote_host}:{self.remote_port}")

        try:
            subprocess.run(cmd)
        except KeyboardInterrupt:
            ...



from functools import lru_cache
from pathlib import Path


@lru_cache()
def get_cache_path() -> Path:
    cache_dir = Path.home() / ".cache" / "webvpn-ng"
    if not cache_dir.exists():
        cache_dir.mkdir(parents=True)

    return cache_dir



from typing import Optional

import json
from functools import lru_cache
from pathlib import Path

import requests
from bs4 import BeautifulSoup

from .utils import get_cache_path

LOGIN_URL_1 = r"https://sso.buaa.edu.cn/login?service=https%3A%2F%2Fd.buaa.edu.cn%2Flogin%3Fcas_login%3Dtrue"
LOGIN_URL_2 = r"https://sso.buaa.edu.cn/login"


@lru_cache()
def get_cookie_cache_path() -> Path:
    return get_cache_path() / "cookie.json"


@lru_cache()
def get_user_cache_path() -> Path:
    return get_cache_path() / "user.json"


def login(username: str, password: str, token: str):
    sess = requests.Session()

    # Step 1: Get `execution` from login page
    rsp = sess.get(LOGIN_URL_1)
    soup = BeautifulSoup(rsp.content, "html.parser")
    tag = soup.find("input", {"name": "execution"})
    assert tag is not None
    execution = tag["value"]

    # Step 2: Post username and password
    data = {
        "username": username,
        "password": password,
        "submit": "登录",
        "type": "username_password",
        "execution": execution,
        "_eventId": "submit",
    }

    rsp = sess.post(LOGIN_URL_2, data=data)

    if rsp.ok:
        cookies = sess.cookies.get_dict()
        with open(get_cookie_cache_path(), "w") as f:
            json.dump(cookies, f)

        with open(get_user_cache_path(), "w") as f:
            json.dump({
                "username": username,
                "password": password,
                "token": token,
            }, f)

    return rsp.ok


def check_login() -> bool:
    if not get_cookie_cache_path().exists():
        return False

    with open(get_cookie_cache_path(), "r") as f:
        cookies = json.load(f)

    sess = requests.Session()
    sess.cookies.update(cookies)

    rsp = sess.get("https://d.buaa.edu.cn/", allow_redirects=False)

    if rsp.status_code == 200:
        return True

    if not get_user_cache_path().exists():
        return False

    with open(get_user_cache_path(), "r") as f:
        user = json.load(f)

    return login(user["username"], user["password"])


def get_cookie() -> str:
    if not get_cookie_cache_path().exists():
        raise RuntimeError("Not logged in")

    with open(get_cookie_cache_path(), "r") as f:
        cookies = json.load(f)

    return "; ".join(f"{k}={v}" for k, v in cookies.items())


def get_token() -> str:
    if not get_user_cache_path().exists():
        raise RuntimeError("Not logged in")

    with open(get_user_cache_path(), "r") as f:
        user = json.load(f)

    return user["token"]


def get_server(token: str) -> Optional[str]:
    rsp = requests.get(
        "https://webvpn.sota.wiki/api/v1/server", params={"token": token}
    )
    data = rsp.json()

    if "server" not in data:
        return None

    return data["server"]



import typer

from .chisel import Chisel
from .login import login as _login, check_login, get_cookie, get_token, get_server

app = typer.Typer(no_args_is_help=True)


@app.command()
def login(
    username: str,
    password: str = typer.Option(..., prompt=True, hide_input=True),
    token: str = typer.Option("", prompt=True),
):
    if len(token) > 0 and ":" not in token:
        typer.echo("Invalid token")
        return

    typer.echo(f"Logging in {username}")

    if _login(username, password, token):
        typer.echo("Login success")
    else:
        typer.echo("Login failed")


@app.command()
def forward(
    host: str = "localhost",
    port: int = "2222",
    rhost: str = "10.251.0.23",
    rport: int = "22",
    delay: int = 100,
    quiet: bool = False,
):
    if not check_login():
        typer.echo("Not logged in")
        return

    typer.echo(f"Forwarding {host}:{port} -> {rhost}:{rport}")

    token = get_token()

    typer.echo("Getting server url... ", nl=False)
    server = get_server(token)
    if not server:
        typer.echo("Failed to get server.")
        return
    else:
        typer.echo("Done.")

    chisel = Chisel(
        server=server,
        cookie=get_cookie(),
        local_host=host,
        local_port=port,
        remote_host=rhost,
        remote_port=rport,
        auth_key=token,
        delay=delay,
        quiet=quiet,
    )
    chisel.run()


if __name__ == "__main__":
    app()
